-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("notifications", "AceEvent-3.0", "AceHook-3.0", "LibSink-2.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local format, gsub, pairs, GetItemInfo = string.format, string.gsub, pairs, GetItemInfo
local origiOnEvent = UIErrorsFrame:GetScript("OnEvent")

-----------------
--- variables ---
-----------------
module.callBacks = {
	["playerQuestComplete"] = true,
	["playerQuestFailed"] = true,
	["partyQuestComplete"] = true,
	["partyQuestFailed"] = true,
	["partyQuestAccepted"] = true,
	["itemStartingQuestLooted"] = true
}
local function paternize(pattern)
	pattern = pattern:gsub("[()]", "%%%1")
	return "^"..pattern:gsub("(%%%d?$?[sd])", "(.+)").."$"
end
local found = paternize(ERR_QUEST_ADD_FOUND_SII)
local item = paternize(ERR_QUEST_ADD_ITEM_SII)
local kill = paternize(ERR_QUEST_ADD_KILL_SII)
local playerKilled = paternize(ERR_QUEST_ADD_PLAYER_KILL_SII)
local oComplete = paternize(ERR_QUEST_OBJECTIVE_COMPLETE_S)
local qComplete = paternize(ERR_QUEST_COMPLETE_S)
local qFailed = paternize(ERR_QUEST_FAILED_S)

-----------------------
--- Tools functions ---
-----------------------
local function normalizeColorComponent(component)
	if component > 1 then return 1 elseif component < 0 then return 0 else return component end
end

local function getColorFromCompletionRGB(percent)
	if percent <= 0 then
		return core.db.profile.quests.rj0, core.db.profile.quests.gj0, core.db.profile.quests.bj0
	elseif percent >= 1 then
		return core.db.profile.quests.rj3, core.db.profile.quests.gj3, core.db.profile.quests.bj3
	elseif percent <= 0.5 then
		return normalizeColorComponent(core.db.profile.quests.ri1*percent + core.db.profile.quests.rj1), normalizeColorComponent(core.db.profile.quests.gi1*percent + core.db.profile.quests.gj1), normalizeColorComponent(core.db.profile.quests.bi1*percent + core.db.profile.quests.bj1)
	else
		return normalizeColorComponent(core.db.profile.quests.ri2*percent + core.db.profile.quests.rj2), normalizeColorComponent(core.db.profile.quests.gi2*percent + core.db.profile.quests.gj2), normalizeColorComponent(core.db.profile.quests.bi2*percent + core.db.profile.quests.bj2)
	end
end

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		playerObjectiveProgress = true,
		playerObjectiveComplete = true,
		playerQuestComplete = true,
		playerQuestFailed = true,
		partyObjectiveProgress = true,
		partyObjectiveComplete = true,
		partyQuestComplete = true,
		partyQuestFailed = true,
		partyQuestAccepted = true,
		partyQuestTurnedIn = true,
		partyQuestAbandonned = true,
		hideBlizzardNotification = true,
		itemStartingQuestLooted = true,
		objectiveNotificationWhenComplete = false,
		sinkOptions = {
		},
	},
}

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("notifications", defaults, "Default")
	self:SetSinkStorage(self.db.profile.sinkOptions)
	self:SetEnabledState(core:getModuleEnabled("notifications"))
	core.moduleState["notifications"] = core:getModuleEnabled("notifications")
end

function module:OnEnable()
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestLost")
	for callback in pairs(self.callBacks) do
		if self.db.profile[callback] then
			core.RegisterCallback(self, callback)
		end
	end
	if self.db.profile.hideBlizzardNotification then
		UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
			if event == "UI_INFO_MESSAGE" and (message:match(found) or message:match(item) or message:match(kill) or message:match(playerKilled) or message:match(oComplete) or message:match(qComplete) or message:match(qFailed)) then return end
			return origiOnEvent(self, event, message, ...)
		end)
	end
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
		return origiOnEvent(self, event, message, ...)
	end)
	self:Refresh()
end

function module:Refresh()
	if not core.moduleState["notifications"] then return end
	core.UnregisterAllCallbacks(self)
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestLost")
	for callback in pairs(self.callBacks) do
		if self.db.profile[callback] then
			core.RegisterCallback(self, callback)
		end
	end
	if self.db.profile.hideBlizzardNotification then
		UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
			if event == "UI_INFO_MESSAGE" and (message:match(found) or message:match(item) or message:match(kill) or message:match(playerKilled) or message:match(oComplete) or message:match(qComplete) or message:match(qFailed)) then return end
			return origiOnEvent(self, event, message, ...)
		end)
	else
		UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
			return origiOnEvent(self, event, message, ...)
		end)
	end
end

function module:SetBlizzNotification()
	if core.db.profile.hideBlizzardNotification then
		UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
			if event == "UI_INFO_MESSAGE" and (message:match(found) or message:match(item) or message:match(kill) or message:match(playerKilled) or message:match(oComplete) or message:match(qComplete) or message:match(qFailed)) then return end
			return origiOnEvent(self, event, message, ...)
		end)
	else
		UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
			return origiOnEvent(self, event, message, ...)
		end)
	end
end

---------------------------
--- Main Module Program ---
---------------------------
function module:playerQuestObjectiveProgress(event, commHandler, uid, objective, got, nGot)
	if nGot == core.questsObjectives[uid][objective].nNeed and self.db.profile.playerObjectiveComplete then
		if not self.db.profile.objectiveNotificationWhenComplete and core.quests[uid].complete[core.playerName] ~= 0 then return end
		self:Pour(format(core.LC["(%s) %s : finished"], core.quests[uid].title, core.questsObjectives[uid][objective].desc), 0, 1, 0)
	elseif self.db.profile.playerObjectiveProgress and nGot > 0 then
		self:Pour("("..core.quests[uid].title..") "..core.questsObjectives[uid][objective].desc.." : "..got.."/"..core.questsObjectives[uid][objective].need, getColorFromCompletionRGB(nGot/core.questsObjectives[uid][objective].nNeed))
	end
end

function module:playerQuestComplete(event, commHandler, uid)
	self:Pour(format(core.LC["(%s) : finished"], core.quests[uid].title), 0, 1, 0)
end

function module:playerQuestFailed(event, commHandler, uid)
	self:Pour(format(core.LC["(%s) : failed"], core.quests[uid].title), 1, 0, 0)
end

function module:partyQuestObjectiveProgress(event, commHandler, uid, objective, got, nGot, sender)
	if nGot == core.questsObjectives[uid][objective].nNeed and self.db.profile.partyObjectiveComplete then
		if not self.db.profile.objectiveNotificationWhenComplete and core.quests[uid].complete[sender] ~= 0 then return end
		self:Pour(format(core.LC["(%s : %s) %s : finished"], sender, core.quests[uid].title, core.questsObjectives[uid][objective].desc), 0, 1, 0)
	elseif self.db.profile.partyObjectiveProgress and nGot > 0 then
		self:Pour("("..sender.." : "..core.quests[uid].title..") "..core.questsObjectives[uid][objective].desc.." : "..got.."/"..core.questsObjectives[uid][objective].need, getColorFromCompletionRGB(nGot/core.questsObjectives[uid][objective].nNeed))
	end
end

function module:partyQuestComplete(event, commHandler, uid, sender)
	self:Pour(format(core.LC["(%s : %s) : finished"], sender, core.quests[uid].title), 0, 1, 0)
end

function module:partyQuestFailed(event, commHandler, uid, sender)
	self:Pour(format(core.LC["(%s : %s) : failed"], sender, core.quests[uid].title), 1, 0, 0)
end

function module:partyQuestLost(event, commHandler, uid, abandoning, title, sender)
	if abandoning and self.db.profile.partyQuestAbandonned then
		self:Pour(format(core.LC["(%s : %s) : abandonned"], sender, title), 1, 0, 0)
	elseif not abandoning and self.db.profile.partyQuestTurnedIn then
		self:Pour(format(core.LC["(%s : %s) : turned in"], sender, title), 0, 1, 0)
	end
end

function module:partyQuestAccepted(event, sender, uid, started, batch)
	if not batch then
		self:Pour(format(core.LC["(%s : %s) : accepted"], sender, core.quests[uid].title), 0, 1, 0)
	end
end

function module:itemStartingQuestLooted(event, itemId)
	local _, itemLink = GetItemInfo(itemId)
	self:Pour(format(core.LC["You've just looted %s which start a quest"], itemLink), 0, 1, 0)
end