-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTracker", "AceEvent-3.0", "AceHook-3.0")
local achievementTracker = core:GetModule("achievementTracker")

-----------------------
--- Library loading ---
-----------------------
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, ipairs, unpack, format, type, sub, tonumber, match, tremove, time, GetTime, floor = pairs, ipairs, unpack, string.format, type, string.sub, tonumber, string.match, tremove, time, GetTime, math.floor
local SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName, WorldMap_OpenToQuest = SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName, WorldMap_OpenToQuest
local SelectQuestLogEntry, QuestLogPushQuest, GetQuestLogSelection, GetQuestLink, GetItemCooldown = SelectQuestLogEntry, QuestLogPushQuest, GetQuestLogSelection, GetQuestLink, GetItemCooldown
local CreateFrame, CreateFont, SetFont, ChatEdit_InsertLink, ItemRefTooltip, ShowUIPanel = CreateFrame, CreateFont, SetFont, ChatEdit_InsertLink, ItemRefTooltip, ShowUIPanel
local GetSubZoneText, GetZoneText, UnitIsDead, GetItemCount, GetQuestLogSpecialItemInfo = GetSubZoneText, GetZoneText, UnitIsDead, GetItemCount, GetQuestLogSpecialItemInfo
local UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown, IsQuestLogSpecialItemInRange = UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown, IsQuestLogSpecialItemInRange

-----------------
--- Variables ---
-----------------
local CELL_MARGIN_H = 6
local minus = "Interface\\Buttons\\UI-MinusButton-Up"
local plus = "Interface\\Buttons\\UI-PlusButton-Up"
local subZone, currentZone, oldZone = "", "", ""
local hTrFontTQ2 = CreateFont("hTrFontTQ2")
local qTrFontTQ2 = CreateFont("qTrFontTQ2")
local qoTrFontTQ2 = CreateFont("qoTrFontTQ2")
local fixQTrFontTQ2 = CreateFont("fixQTrFontTQ2")
local backdrop = GameTooltip:GetBackdrop()
local rangeUpdateDelay = TOOLTIP_UPDATE_TIME
local MAX_PLAYER_LEVEL = MAX_PLAYER_LEVEL
local autoUntrack, delayedHide = {}, {}
local anchorSetPoint = {
	[1] = {point = "TOPLEFT", relPoint = "BOTTOMLEFT", x = 0, y = 24},
	[2] = {point = "TOP", relPoint = "BOTTOM", x = 0, y = 24},
	[3] = {point = "TOPRIGHT", relPoint = "BOTTOMRIGHT", x = 0, y = 24},
	[4] = {point = "BOTTOMLEFT", relPoint = "TOPLEFT", x = 0, y = -24},
	[5] = {point = "BOTTOM", relPoint = "TOP", x = 0, y = -24},
	[6] = {point = "BOTTOMRIGHT", relPoint = "TOPRIGHT", x = 0, y = -24}}

-----------------------
--- Tools functions ---
-----------------------
local function normalizeWidth(min, max)
	if max < min then return max else return min end
end

local function normalizeColorComponent(component)
	if component > 1 then return 1 elseif component < 0 then return 0 else return component end
end

local function formatTime(s)
	if s >= 60 then
		return floor(s/60+0.5).."m"..floor(s%60).."s"
	else
		return s.."s"
	end
end

local function getColorFromCompletionHEX(percent)
	if percent <= 0 then
		return format("%02x%02x%02x", 255*core.db.profile.quests.rj0, 255*core.db.profile.quests.gj0, 255*core.db.profile.quests.bj0)
	elseif percent >= 1 then
		return format("%02x%02x%02x", 255*core.db.profile.quests.rj3, 255*core.db.profile.quests.gj3, 255*core.db.profile.quests.bj3)
	elseif percent <= 0.5 then
		return format("%02x%02x%02x", 255*normalizeColorComponent(core.db.profile.quests.ri1*percent + core.db.profile.quests.rj1), 255*normalizeColorComponent(core.db.profile.quests.gi1*percent + core.db.profile.quests.gj1), 255*normalizeColorComponent(core.db.profile.quests.bi1*percent + core.db.profile.quests.bj1))
	else
		return format("%02x%02x%02x", 255*normalizeColorComponent(core.db.profile.quests.ri2*percent + core.db.profile.quests.rj2), 255*normalizeColorComponent(core.db.profile.quests.gi2*percent + core.db.profile.quests.gj2), 255*normalizeColorComponent(core.db.profile.quests.bi2*percent + core.db.profile.quests.bj2))
	end
end

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		onLogIn = false,
		onQuestAccepted = 3,
		onObjectiveProgress = true,
		unTrackInactive = false,
		unTrackDelay = 5,
		onSubZone = true,
		onZone = true,
		onDaily = true,
		onClass = true,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedQuest = 2,
		delayToHideCompletedQuest = 0,
		hideCompletedObjective = 2,
		onlyCurrentZone = false,
		limitQuestLevel = false,
		bgTexture = "Blizzard Tooltip",
		borderTexture = "Blizzard Tooltip",
		borderInset = 2.5,
		borderWidth = 14,
		headerFont = LSM.DefaultMedia.font,
		headerFontSize = 12,
		headerFontOutline = "",
		questFont = LSM.DefaultMedia.font,
		questFontSize = 10,
		questFontOutline = "",
		questObjectiveFont = LSM.DefaultMedia.font,
		questObjectiveFontSize = 8,
		questObjectiveFontOutline = "",
		maxWidth = 250,
		minWidth = 250,
		statusMaxWidth = 70,
		statusMinWidth = 70,
		hasMaxHeight = true,
		maxHeight = 600,
		scale = 1,
		showQuestItemIcon = true,
		oorColor = {0.8, 0.1, 0.1},
		questItemIconSize = 20,
		showItemsStartingQuest = true,
		otherItemIconSize = 12,
		showUsableQuestItemsNotAttached = true,
		questsCellPadding = 5,
		objectivesCellPadding = 10,
		rightCellPadding = 5,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		colorizeQuestName = true,
		defaultQuestNameColor = "ffffff",
		colorizeQuestLevel = true,
		defaultQuestLevelColor = "ffffff",
		colorizeQuestObjective = true,
		defaultQuestObjectiveColor = "ffffff",
		colorizeQuestObjectiveCounter = true,
		colorizeQuestTimer = true,
		defaultQuestTimerColor = "ffffff",
		defaultQuestDescriptionColor = "ffffff",
		displayHeader = true,
		autoCollapseHeader = true,
		showCollapseButton = true,
		collapseButtonSize = 13,
		colorizeHeaderName = true,
		defaultHeaderNameColor = "ffffff",
		hideZoneLevelRange = false,
		hideQuestTag = false,
		colorizeHeaderLevel = true,
		defaultHeaderLevelColor = "ffffff",
		disableInteraction = false,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "objectiveProgress",
		mouseRightAlt = "macroUpdate",
		mouseRight = "questTooltip",
		mouseLeftShift = "trackQuest",
		mouseLeftControl = "shareQuest",
		mouseLeftAlt = "abandonQuest",
		mouseLeft = "displayQuest",
		mouseMiddleShift = "displayQuestInfo",
		mouseMiddleControl = "displayQuestInfo",
		mouseMiddleAlt = "displayQuestInfo",
		mouseMiddle = "displayQuestInfo",
	},
}

-------------------
--- Timed Stuff ---
-------------------
local function checkTimer(self, elapsed)
	self.duration = self.duration + elapsed
	if self.duration >= 1 then
		self.currentTime = self.currentTime + self.duration
		self.duration = 0
		local updateTracker
		for uid, endTime in pairs(autoUntrack) do
			if self.currentTime >= endTime then
				autoUntrack[uid] = nil
				RemoveQuestWatch(core.quests[uid].id, uid, true)
				updateTracker = true
			end
		end
		for uid, endTime in pairs(delayedHide) do
			if self.currentTime >= endTime then
				delayedHide[uid] = nil
				updateTracker = true
			end
		end
		if not next(autoUntrack) and not next(delayedHide) then
			self:Hide()
		end
		if updateTracker then
			module:updateQuestsTracker()
		end
	end
end
local timedQuestTrackerUpdate = CreateFrame("frame", "tomQuest2TimedQuestTrackerUpdate")
timedQuestTrackerUpdate.duration = 0
timedQuestTrackerUpdate.currentTime = 0
timedQuestTrackerUpdate:SetScript("OnUpdate", checkTimer)
timedQuestTrackerUpdate:Hide()

----------------------
--- Initialization ---
----------------------
local disableInteraction, collapseButtonSize, maxWidth, statusMaxWidth, minWidth, statusMinWidth, questItemIconSize, questsCellPadding, objectivesCellPadding, totalMaxWidth, totalMinWidth, numColumns
function module:initTrackerSettings()
	disableInteraction = self.db.profile.disableInteraction
	collapseButtonSize = self.db.profile.showCollapseButton and self.db.profile.collapseButtonSize or 0
	maxWidth = self.db.profile.maxWidth
	statusMaxWidth = self.db.profile.statusMaxWidth
	minWidth = normalizeWidth(self.db.profile.minWidth, maxWidth)
	statusMinWidth = normalizeWidth(self.db.profile.statusMinWidth, statusMaxWidth)
	if self.db.profile.showQuestItemIcon then
		questItemIconSize = self.db.profile.questItemIconSize
		questsCellPadding = self.db.profile.questsCellPadding + questItemIconSize
		objectivesCellPadding = self.db.profile.objectivesCellPadding + questItemIconSize
	else
		questItemIconSize = 0
		questsCellPadding = self.db.profile.questsCellPadding
		objectivesCellPadding = self.db.profile.objectivesCellPadding
	end
	if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
		achievementTracker:initTrackerSettings(disableInteraction, maxWidth, statusMaxWidth, minWidth, statusMinWidth)
		core.cbh:Fire("updateAchievementTracker")
	end
end

local function setBackDrop()
	backdrop.bgFile = LSM:Fetch("background", module.db.profile.bgTexture)
	backdrop.insets.left = module.db.profile.borderInset
	backdrop.insets.right = module.db.profile.borderInset
	backdrop.insets.top = module.db.profile.borderInset
	backdrop.insets.bottom = module.db.profile.borderInset
	backdrop.edgeFile = LSM:Fetch("border", module.db.profile.borderTexture)
	backdrop.edgeSize = module.db.profile.borderWidth
end
module.setBackDrop = setBackDrop

local function setFonts()
	hTrFontTQ2:SetFont(LSM:Fetch("font", module.db.profile.headerFont), module.db.profile.headerFontSize, module.db.profile.headerFontOutline)
	qTrFontTQ2:SetFont(LSM:Fetch("font", module.db.profile.questFont), module.db.profile.questFontSize, module.db.profile.questFontOutline)
	qoTrFontTQ2:SetFont(LSM:Fetch("font", module.db.profile.questObjectiveFont), module.db.profile.questObjectiveFontSize, module.db.profile.questObjectiveFontOutline)
end
module.setFonts = setFonts

local function setTrackerInCombat()
	if module.db.profile.hideTrackerInCombat then
		module:RegisterEvent("PLAYER_REGEN_ENABLED", "outOfCombat")
		module:RegisterEvent("PLAYER_REGEN_DISABLED", "inCombat")
	else
		module:RegisterEvent("PLAYER_REGEN_ENABLED", "outOfCombat")
		module:RegisterEvent("PLAYER_REGEN_DISABLED", "inCombat")
	end
end
module.setTrackerInCombat = setTrackerInCombat

function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTracker", defaults, "Default")
	self.db.char.questsTracked = self.db.char.questsTracked or {}
	self.db.char.questsSubZoneAutoTracked = self.db.char.questsSubZoneAutoTracked or {}
	self.db.char.questsZoneAutoTracked = self.db.char.questsZoneAutoTracked or {}
	self.db.char.zoneDisplay = self.db.char.zoneDisplay or {}
	self:SetEnabledState(core:getModuleEnabled("questsTracker"))
	core.moduleState["questsTracker"] = core:getModuleEnabled("questsTracker")
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
end

function module:OnEnable()
	setBackDrop()
	setFonts()
	fixQTrFontTQ2:SetFont(LSM:Fetch("font", LSM.DefaultMedia.font), 1)
	subZone = GetSubZoneText() or ""
	currentZone = GetZoneText() or ""
	self.anchor = CreateFrame("Button", nil, UIParent)
	self.anchor:SetHeight(24)
	self.anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	self.anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
	self.anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
	self.anchor.text = self.anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
	self.anchor.text:SetPoint("CENTER")
	self.anchor.text:SetText(core.LC["Quests Tracker"])
	self.anchor:SetWidth(self.anchor.text:GetStringWidth() + 8)
	self.anchor:SetMovable(true)
	self.anchor:RegisterForDrag("LeftButton")
	self.anchor:SetClampedToScreen(true)
	self.anchor:SetScript("OnDragStart", function(self) self:StartMoving() end)
	self.anchor:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		module.db.profile.point, module.db.profile.x, module.db.profile.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
		module.trackerAnchor = module.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
		module:updateQuestsTracker()
	end)
	self.anchor:SetPoint(self.db.profile.point, self.db.profile.x, self.db.profile.y)
	if not self.db.profile.showAnchor then self.anchor:Hide() else self.anchor:Show() end
	self.trackerAnchor = self.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	self:initTrackerSettings()
	setTrackerInCombat()
	self:RegisterEvent("ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "zoneChanged")
	self:RawHook("IsQuestWatched", true)
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	core.RegisterCallback(self, "init")
	core.RegisterCallback(self, "playerQuestAccepted")
	core.RegisterCallback(self, "partyQuestAccepted")
	core.RegisterCallback(self, "playerQuestComplete", "playerQuestComplete")
	core.RegisterCallback(self, "playerQuestUnComplete", "playerQuestUnComplete")
	core.RegisterCallback(self, "playerQuestFailed", "questFailed")
	core.RegisterCallback(self, "partyQuestComplete", "partyQuestComplete")
	core.RegisterCallback(self, "partyQuestUnComplete", "partyQuestUnComplete")
	core.RegisterCallback(self, "partyQuestFailed", "questFailed")
	core.RegisterCallback(self, "playerQuestObjectiveProgress", "questObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress", "questObjectiveProgress")
	core.RegisterCallback(self, "playerQuestLost", "questLost")
	core.RegisterCallback(self, "partyQuestLost", "questLost")
	core.RegisterCallback(self, "toggleTrackedQuest")
	core.RegisterCallback(self, "questObjectiveAdded")
	core.RegisterCallback(self, "questDescriptionAdded")
	core.RegisterCallback(self, "updateQuestsTracker")
end

function module:OnDisable()
	self.anchor:Hide()
	QTIP:Release(self.tracker)
	self.tracker = nil
	core.UnregisterAllCallbacks(self)
	if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
		achievementTracker:initTrackerSettings(achievementTracker.db.profile.disableInteraction, achievementTracker.db.profile.maxWidth, achievementTracker.db.profile.statusMaxWidth, achievementTracker.db.profile.minWidth, achievementTracker.db.profile.statusMinWidth)
		core.cbh:Fire("updateAchievementTracker")
	end
	self:Refresh()
end

function module:Refresh()
	if not core.moduleState["questsTracker"] then return end
	self:UnregisterAllEvents()
	setFonts()
	fixQTrFontTQ2:SetFont(LSM:Fetch("font", LSM.DefaultMedia.font), 1)
	self.anchor:SetPoint(self.db.profile.point, self.db.profile.x, self.db.profile.y)
	if not self.db.profile.showAnchor then self.anchor:Hide() else self.anchor:Show() end
	self.trackerAnchor = self.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	self:initTrackerSettings()
	setTrackerInCombat()
end

function module:updateUsedMedia(event, mediatype, key)
	if not core.moduleState["questsTracker"] then return end
	if mediatype == "font" then
		local updateQuestTracker
		if key == self.db.profile.headerFont then
			updateQuestTracker = true
			hTrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.headerFont), self.db.profile.headerFontSize, self.db.profile.headerFontOutline)
		end
		if key == self.db.profile.questFont then
			updateQuestTracker = true
			qTrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.questFont), self.db.profile.questFontSize, self.db.profile.questFontOutline)
		end
		if key == self.db.profile.questObjectiveFont then
			updateQuestTracker = true
			qoTrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.questObjectiveFont), self.db.profile.questObjectiveFontSize, self.db.profile.questObjectiveFontOutline)
		end
		if updateQuestTracker then
			self:updateQuestsTracker()
		end
	end
end

-------------------
--- zoneChanged ---
-------------------
function module:zoneChanged()
	if UnitIsDead("player") then return end
	local trackerNeedUpdate
	subZone = GetSubZoneText() or ""
	currentZone = GetZoneText() or ""
	if self.db.profile.onZone and currentZone ~= oldZone then
		for uid in pairs(self.db.char.questsZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				RemoveQuestWatch(core.quests[uid].id, uid, true, true)
			else
				self.db.char.questsZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not self.db.char.questsTracked[uid] and qData.zone == currentZone then
				self.db.char.questsZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				AddQuestWatch(core.quests[uid].id, nil, uid, true)
			end
		end
	end
	if self.db.profile.onSubZone and subZone ~= "" then
		for uid in pairs(self.db.char.questsSubZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				RemoveQuestWatch(core.quests[uid].id, uid, true, true)
			else
				self.db.char.questsSubZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not self.db.char.questsTracked[uid] and ((qData.objective and qData.objective:find(subZone, 1, true)) or (qData.description and qData.description:find(subZone, 1, true))) then
				self.db.char.questsSubZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				AddQuestWatch(core.quests[uid].id, nil, uid, true)
			end
		end
	end
	if self.db.profile.displayHeader and self.db.profile.autoCollapseHeader then
		for zone, zData in pairs(core.zones) do
			if currentZone == zone then
				self.db.char.zoneDisplay[zone] = 1
			elseif zData.zone then
				self.db.char.zoneDisplay[zone] = 0
			end
		end
		trackerNeedUpdate = true
	end
	if trackerNeedUpdate then
		for uid in pairs(autoUntrack) do
			if not self.db.char.questsTracked[uid] then
				autoUntrack[uid] = nil
			end
		end
		for uid in pairs(delayedHide) do
			if not self.db.char.questsTracked[uid] then
				delayedHide[uid] = nil
			end
		end
		self:updateQuestsTracker()
	end
	oldZone = currentZone
end

-------------
--- Hooks ---
-------------
function module:IsQuestWatched(id)
	local uid = tonumber(match(GetQuestLink(id) or '', 'quest:(%d+)'))
	return self.db.char.questsTracked[uid] and 1 or nil
end

function module:AddQuestWatch(id, timeToTrack, uid, batch)
	if not uid then
		uid = tonumber(match(GetQuestLink(id) or '', 'quest:(%d+)'))
		if not uid or not core.quests[uid] then return end
	end
	self.db.char.questsTracked[uid] = true
	core.quests[uid].tracked = true
	if not batch then
		self:updateQuestsTracker()
	end
	if QuestLogFrame:IsVisible() then
		QuestLog_Update()
	end
end

function module:RemoveQuestWatch(id, uid, batch)
	if not uid then
		uid = tonumber(match(GetQuestLink(id) or '', 'quest:(%d+)'))
		if not uid or not core.quests[uid] then return end
	end
	if self.db.char.questsTracked[uid] then
		self.db.char.questsTracked[uid] = nil
		self.db.char.questsSubZoneAutoTracked[uid] = nil
		self.db.char.questsZoneAutoTracked[uid] = nil
		core.quests[uid].tracked = nil
		if not batch then
			if autoUntrack[uid] or delayedHide[uid] then
				autoUntrack[uid] = nil
				delayedHide[uid] = nil
				if not next(autoUntrack) and not next(delayedHide) then
					timedQuestTrackerUpdate:Hide()
					timedQuestTrackerUpdate.duration = 0
				end
			end
			self:updateQuestsTracker()
		end
	end
	if QuestLogFrame:IsVisible() then
		QuestLog_Update()
	end
end

-----------------
--- Callbacks ---
-----------------
function module:init(event)
	if self.db.profile.onLogIn then
		for uid in pairs (core.quests) do
			AddQuestWatch(core.quests[uid].id, nil, uid, true)
		end
	else
		for uid in pairs(self.db.char.questsTracked) do
			if not core.quests[uid] then
				self.db.char.questsTracked[uid] = nil
				self.db.char.questsSubZoneAutoTracked[uid] = nil
				self.db.char.questsZoneAutoTracked[uid] = nil
			else
				AddQuestWatch(core.quests[uid].id, nil, uid, true)
			end
		end
	end
	self:updateQuestsTracker()
end

function module:playerQuestAccepted(event, commHandler, uid, started)
	if self.db.profile.onQuestAccepted == 1 or (self.db.profile.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) or (self.db.profile.onObjectiveProgress and started == 1) or (self.db.profile.onDaily and core.quests[uid].daily == 1) or (self.db.profile.onClass and core.quests[uid].class) then
		if not self.db.char.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
		else
			if autoUntrack[uid] and (self.db.profile.onQuestAccepted == 1 or (self.db.profile.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) or (self.db.profile.onDaily and core.quests[uid].daily == 1) or (self.db.profile.onClass and core.quests[uid].class)) then
				autoUntrack[uid] = nil
				if not next(autoUntrack) and not next(delayedHide) then
					timedQuestTrackerUpdate:Hide()
					timedQuestTrackerUpdate.duration = 0
				end
			end
			self.db.char.questsSubZoneAutoTracked[uid] = nil
			self.db.char.questsZoneAutoTracked[uid] = nil
			self:updateQuestsTracker()
		end
	elseif self.db.profile.onZone and core.quests[uid].zone == currentZone then
		if not self.db.char.questsTracked[uid] then
			self.db.char.questsZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		else
			self.db.char.questsSubZoneAutoTracked[uid] = nil
			self:updateQuestsTracker()
		end
	elseif self.db.profile.onSubZone and not self.db.char.questsTracked[uid] and subZone ~= "" and ((core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true)) or (core.quests[uid].description and core.quests[uid].description:find(subZone, 1, true))) then
		self.db.char.questsSubZoneAutoTracked[uid] = true
		AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
	elseif self.db.char.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:partyQuestAccepted(event, sender, uid, started, batch)
	if batch then
		if not self.db.char.questsTracked[uid] and self.db.profile.onLogIn then
			AddQuestWatch(core.quests[uid].id, nil, uid, true)
		end
	elseif self.db.profile.onQuestAccepted == 1 or (self.db.profile.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) or (self.db.profile.onObjectiveProgress and started == 1) or (self.db.profile.onDaily and core.quests[uid].daily == 1) then
		if not self.db.char.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
		else
			if autoUntrack[uid] and (self.db.profile.onQuestAccepted == 1 or (self.db.profile.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) or (self.db.profile.onDaily and core.quests[uid].daily == 1)) then
				autoUntrack[uid] = nil
				if not next(autoUntrack) and not next(delayedHide) then
					timedQuestTrackerUpdate:Hide()
					timedQuestTrackerUpdate.duration = 0
				end
			end
			self.db.char.questsSubZoneAutoTracked[uid] = nil
			self.db.char.questsZoneAutoTracked[uid] = nil
			self:updateQuestsTracker()
		end
	elseif self.db.char.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:playerQuestComplete(event, commHandler, uid, sender)
	local completeTime = time()
	core.quests[uid].playerCompleteTime = completeTime
	if core.quests[uid].partyComplete == 1 then
		core.quests[uid].partyCompleteTime = completeTime
	end
	if self.db.char.questsTracked[uid] then
		if self.db.profile.delayToHideCompletedQuest > 0 and ((self.db.profile.hideCompleteQuests == 1 and core.quests[uid].playerComplete == 1) or (self.db.profile.hideCompleteQuests == 2 and core.quests[uid].partyComplete == 1)) then
			delayedHide[uid] = completeTime + self.db.profile.delayToHideCompletedQuest
			if not timedQuestTrackerUpdate:IsShown() then
				timedQuestTrackerUpdate.currentTime = time()
				timedQuestTrackerUpdate:Show()
			end
		end
		self:updateQuestsTracker()
	end
	if autoUntrack[uid] then
		autoUntrack[uid] = nil
		if not next(autoUntrack) and not next(delayedHide) then
			timedQuestTrackerUpdate:Hide()
			timedQuestTrackerUpdate.duration = 0
		end
	end
end

function module:partyQuestComplete(event, commHandler, uid, sender)
	if core.quests[uid].partyComplete == 1 then
		local completeTime = time()
		core.quests[uid].partyCompleteTime = completeTime
		if self.db.char.questsTracked[uid] then
			if self.db.profile.delayToHideCompletedQuest > 0 and self.db.profile.hideCompleteQuests == 2 then
				delayedHide[uid] = completeTime + self.db.profile.delayToHideCompletedQuest
				if not timedQuestTrackerUpdate:IsShown() then
					timedQuestTrackerUpdate.currentTime = time()
					timedQuestTrackerUpdate:Show()
				end
			end
			self:updateQuestsTracker()
		end
		if autoUntrack[uid] then
			autoUntrack[uid] = nil
			if not next(autoUntrack) and not next(delayedHide) then
				timedQuestTrackerUpdate:Hide()
				timedQuestTrackerUpdate.duration = 0
			end
		end
	elseif self.db.char.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:playerQuestUnComplete(event, commHandler, uid, sender)
	core.quests[uid].playerCompleteTime = nil
	core.quests[uid].partyCompleteTime = nil
	if delayedHide[uid] then
		delayedHide[uid] = nil
		if not next(autoUntrack) and not next(delayedHide) then
			timedQuestTrackerUpdate:Hide()
			timedQuestTrackerUpdate.duration = 0
		end
	end
	if self.db.char.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:partyQuestUnComplete(event, commHandler, uid, sender)
	core.quests[uid].partyCompleteTime = nil
	if delayedHide[uid] and self.db.profile.hideCompleteQuests == 2 then
		delayedHide[uid] = nil
		if not next(autoUntrack) and not next(delayedHide) then
			timedQuestTrackerUpdate:Hide()
			timedQuestTrackerUpdate.duration = 0
		end
	end
	if self.db.char.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:questFailed(event, commHandler, uid, sender)
	if self.db.char.questsTracked[uid] then
		if autoUntrack[uid] then
			autoUntrack[uid] = nil
			if not next(autoUntrack) and not next(delayedHide) then
				timedQuestTrackerUpdate:Hide()
				timedQuestTrackerUpdate.duration = 0
			end
		end
		self:updateQuestsTracker()
	end
end

function module:questObjectiveProgress(event, commHandler, uid, objective, got, nGot, sender)
	if self.db.profile.onObjectiveProgress then
		if not self.db.char.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
			if self.db.profile.unTrackInactive and self.db.profile.unTrackDelay > 0 then
				autoUntrack[uid] = time() + self.db.profile.unTrackDelay * 60
				if not timedQuestTrackerUpdate:IsShown() then
					timedQuestTrackerUpdate.currentTime = time()
					timedQuestTrackerUpdate:Show()
				end
			end
		else
			self.db.char.questsSubZoneAutoTracked[uid] = nil
			self.db.char.questsZoneAutoTracked[uid] = nil
			if self.db.profile.unTrackInactive and self.db.profile.unTrackDelay > 0 and autoUntrack[uid] then
				autoUntrack[uid] = time() + self.db.profile.unTrackDelay * 60
			end
		end
		self:updateQuestsTracker()
	elseif self.db.char.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:questLost(event, commHandler, uid, abandoning, title, sender)
	if not core.quests[uid] and self.db.char.questsTracked[uid] then
		if autoUntrack[uid] or delayedHide[uid] then
			autoUntrack[uid] = nil
			delayedHide[uid] = nil
			if not next(autoUntrack) and not next(delayedHide) then
				timedQuestTrackerUpdate:Hide()
				timedQuestTrackerUpdate.duration = 0
			end
		end
		self.db.char.questsTracked[uid] = nil
		self.db.char.questsSubZoneAutoTracked[uid] = nil
		self.db.char.questsZoneAutoTracked[uid] = nil
		self:updateQuestsTracker()
	elseif core.quests[uid] and self.db.char.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:toggleTrackedQuest(event, uid)
	if self.db.char.questsTracked[uid] then
		RemoveQuestWatch(core.quests[uid].id, uid)
	elseif not self.db.char.questsTracked[uid] then
		AddQuestWatch(core.quests[uid].id, nil, uid)
	end
end

function module:questObjectiveAdded(event, uid)
	if self.db.profile.onSubZone and self.subZone ~= "" then
		if not self.db.char.questsTracked[uid] and core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true) then
			self.db.char.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		elseif core.quests[uid].numObjectives == 0 then
			self:updateQuestsTracker()
		end
	elseif core.quests[uid].numObjectives == 0 then
		self:updateQuestsTracker()
	end
end

function module:questDescriptionAdded(event, uid)
	if self.db.profile.onSubZone and subZone ~= "" then
		if not self.db.char.questsTracked[uid] and core.quests[uid].description and core.quests[uid].description:find(subZone, 1, true) then
			self.db.char.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		end
	end
end

function module:inCombat(event)
	self.hideInCombat = true
	if self.tracker then self.tracker:Hide() end
end

function module:outOfCombat(event)
	self.hideInCombat = nil
	self:updateQuestsTracker()
end

---------------
--- Tracker ---
---------------
function module:ReleaseTracker()
	QTIP:Release(self.tracker)
	self.tracker = nil
end

local interactivity = {
	chatLink = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ChatFrameEditBox:Show()
			ChatEdit_InsertLink(GetQuestLink(core.quests[cValue1].id))
		end
	end,
	objectiveProgress = function(cValue1)
		if core.quests[cValue1].numObjectives > 0 and core.quests[cValue1].complete[core.playerName] then
			ChatFrameEditBox:Show()
			for objective, qoData in pairs(core.questsObjectives[cValue1]) do
				ChatFrameEditBox:Insert("{"..qoData.desc.." "..qoData.got[core.playerName].."/"..qoData.need.."}")
			end
		end
	end,
	macroUpdate = function(cValue1)
		core:generateSpecialItemMacro(cValue1)
	end,
	questTooltip = function(cValue1)
		ItemRefTooltip:ClearLines()
		ShowUIPanel(ItemRefTooltip)
		if not ItemRefTooltip:IsVisible() then
			ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
		end
		ItemRefTooltip:SetHyperlink("quest:"..cValue1)
	end,
	shareQuest = function(cValue1)
		if core.numPartyMembers > 0 and core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id and core.quests[cValue1].sharable then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			QuestLogPushQuest()
			SelectQuestLogEntry(wasSelected)
		end
	end,
	trackQuest = function(cValue1)
		module:toggleTrackedQuest(nil, cValue1)
	end,
	abandonQuest = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			SetAbandonQuest()
			local items = GetAbandonQuestItems()
			if items then
				StaticPopup_Hide("ABANDON_QUEST")
				StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
			else
				StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
				StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
			end
			SelectQuestLogEntry(wasSelected)
		end
	end,
	displayQuest = function(cValue1)
		core:displayQuestInformations(cValue1)
	end,
	displayQuestInfo = function(cValue1)
		WorldMap_OpenToQuest(cValue1)
	end,
	nothing = function(cValue1)
		return
	end
}

local function zoneOnMouseDown(cell, button)
	local cValue1 = cell.cValue1
	if module.db.char.zoneDisplay[cValue1] == 1 then
		module.db.char.zoneDisplay[cValue1] = 0
	else
		module.db.char.zoneDisplay[cValue1] = 1
	end
	module:updateQuestsTracker()
end

local function questOnMouseDown(cell, button)
	local cValue1 = cell.cValue1
	if button == "RightButton" then
		if IsShiftKeyDown() then
			interactivity[module.db.profile.mouseRightShift](cValue1)
		elseif IsControlKeyDown() then
			interactivity[module.db.profile.mouseRightControl](cValue1)
		elseif IsAltKeyDown() then
			interactivity[module.db.profile.mouseRightAlt](cValue1)
		else
			interactivity[module.db.profile.mouseRight](cValue1)
		end
	elseif button == "LeftButton" then
		if IsShiftKeyDown() then
			interactivity[module.db.profile.mouseLeftShift](cValue1)
		elseif IsControlKeyDown() then
			interactivity[module.db.profile.mouseLeftControl](cValue1)
		elseif IsAltKeyDown() then
			interactivity[module.db.profile.mouseLeftAlt](cValue1)
		else
			interactivity[module.db.profile.mouseLeft](cValue1)
		end
	elseif button == "MiddleButton" then
		if IsShiftKeyDown() then
			interactivity[module.db.profile.mouseMiddleShift](cValue1)
		elseif IsControlKeyDown() then
			interactivity[module.db.profile.mouseMiddleControl](cValue1)
		elseif IsAltKeyDown() then
			interactivity[module.db.profile.mouseMiddleAlt](cValue1)
		else
			interactivity[module.db.profile.mouseMiddle](cValue1)
		end
	end
end

local function questItemOnMouseDown(cell, button)
	local cValue1 = cell.cValue1
	core:generateSpecialItemMacro(nil, cValue1)
end

local function questOnEnter(cell)
	if module.db.profile.overTooltip then
		GameTooltip:SetOwner(module.tracker, module.trackerAnchor)
		GameTooltip:SetHyperlink("quest:"..cell.cValue1)
	end
end

local function questOnLeave(cell)
	GameTooltip:Hide()
end

local function buttonOnEnter(self)
	local link = self.link
	if not link then return end
	local _, itemLink = GetItemInfo(link)
	GameTooltip:SetOwner(self, "ANCHOR_LEFT")
	GameTooltip:SetHyperlink(itemLink)
end

local function buttonOnLeave()
	GameTooltip:Hide()
end

local function buttonOnClick(button)
	local category = button.category
	if not category then return end
	if module.db.char.zoneDisplay[category] == 1 then
		module.db.char.zoneDisplay[category] = 0
	else
		module.db.char.zoneDisplay[category] = 1
	end
	module:updateQuestsTracker()
end

local function wrongButtonOnClick(button)
	core:Print(core.LC["This button is not clickable, you can use tomQuest2 macro system instead"])
end

local function updateItemRange(self, elapsed)
	self.rangeTimer = self.rangeTimer - elapsed
	if self.rangeTimer <= 0  then
		if IsQuestLogSpecialItemInRange(self:GetID()) == 0 then
			self.iconTexture:SetVertexColor(0.8, 0.1, 0.1)
		else
			self.iconTexture:SetVertexColor(1.0, 1.0, 1.0)
		end
		self.rangeTimer = rangeUpdateDelay
	end
end

local function itemOnEvent(self, event, ...)
	if event == "BAG_UPDATE_COOLDOWN" then
		local start, duration, enable = GetItemCooldown(self.link)
		if start > 0 and duration > 0 and enable > 0 then
			self.coolDown:SetCooldown(start, duration)
			self.coolDown:Show()
		else
			self.coolDown:Hide()
		end
	elseif event == "PLAYER_TARGET_CHANGED" then
		self.rangeTimer = -1
	end
end

local function timedQuestUpdate(self, elapsed)
	self.duration = self.duration + elapsed
	if self.duration >= 1 then
		local uid = self.cValue1
		if core.quests[uid] and core.quests[uid].remainingTime then
			core.quests[uid].remainingTime = core.quests[uid].remainingTime - self.duration
			self.duration = 0
			if core.quests[uid].playersOnQuest > 1 then
				self.fontString:SetText("|cff"..(module.db.profile.colorizeQuestLevel and core.quests[uid].color or module.db.profile.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(module.db.profile.colorizeQuestName and core.quests[uid].color or module.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest.."> |cff"..(module.db.profile.colorizeQuestTimer and getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or module.db.profile.defaultQuestTimerColor)..formatTime(core.quests[uid].remainingTime).."|r")
			else
				self.fontString:SetText("|cff"..(module.db.profile.colorizeQuestLevel and core.quests[uid].color or module.db.profile.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(module.db.profile.colorizeQuestName and core.quests[uid].color or module.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r |cff"..(module.db.profile.colorizeQuestTimer and getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or module.db.profile.defaultQuestTimerColor)..formatTime(core.quests[uid].remainingTime).."|r")
			end
		else
			self:SetScript("OnUpdate", nil)
		end
	end
end

local zProvider, zPrototype, baseCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
local qProvider, qPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
local qtProvider, qtPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
local qoProvider, qoPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
local qiProvider, qiPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)

function zPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cValue1, cValue2)
	local width, height = baseCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	self:EnableMouse(true)
	self.cValue1 = cValue1
	self:SetScript("OnMouseDown", zoneOnMouseDown)
	if module.db.profile.showCollapseButton then
		self.button = core:getCollapseButton()
		self.button:RegisterForClicks("AnyDown")
		self.button.iconTexture:SetTexture(cValue2)
		self.button:SetScript("OnClick", buttonOnClick)
		self.button.category = cValue1
		if collapseButtonSize > height then
			self.button:SetHeight(height)
			self.button:SetWidth(height)
		else
			self.button:SetHeight(collapseButtonSize)
			self.button:SetWidth(collapseButtonSize)
		end
		self.button:SetParent(self)
		self.button:SetPoint("TOPLEFT", self, "TOPLEFT")
		self.button:SetToplevel(true)
		self.button:Show()
	end
	return width, height
end

function zPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	end
	self:EnableMouse(false)
	self.cValue1 = nil
	self:SetScript("OnMouseDown", nil)
end

function qPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cValue1, cValue2)
	local width, height = baseCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	self:EnableMouse(true)
	self.cValue1 = cValue1
	self:SetScript("OnEnter", questOnEnter)
	self:SetScript("OnLeave", questOnLeave)
	self:SetScript("OnMouseDown", questOnMouseDown)
	if cValue2 ~= 0 then
		if module.db.profile.showQuestItemIcon and (module.db.profile.hideCompletedObjective == 3 or (module.db.profile.hideCompletedObjective == 2 and core.quests[cValue1].partyComplete == 0) or (module.db.profile.hideCompletedObjective == 1 and core.quests[cValue1].playerComplete == 0)) then
			if core.quests[cValue1].item and core.quests[cValue1].id then
				self.button = core:getItemButton()
				self.button:SetID(core.quests[cValue1].id)
				self.button:SetScript("OnEvent", itemOnEvent)
				local itemCount = GetItemCount(cValue2)
				if itemCount > 1 then
					self.button.count:SetText(itemCount)
					self.button.count:Show()
				end
				if core.quests[cValue1].itemHasRange then
					self.button.rangeTimer = -1
					self.button:SetScript("OnUpdate", updateItemRange)
					self.button:RegisterEvent("PLAYER_TARGET_CHANGED")
				end
				self.button:RegisterEvent("BAG_UPDATE_COOLDOWN")
				local start, duration, enable = GetItemCooldown(cValue2)
				if start > 0 and duration > 0 and enable > 0 then
					self.button.coolDown:SetCooldown(start, duration)
					self.button.coolDown:Show()
				else
					self.button.coolDown:Hide()
				end
			else
				self.button = core:getNormalButton()
				self.button:SetScript("OnEnter", buttonOnEnter)
				self.button:SetScript("OnLeave", buttonOnLeave)
				self.button:SetScript("OnClick", wrongButtonOnClick)
			end
			self.button.iconTexture:SetTexture(core.quests[cValue1].specialItemIcon)
			self.button.link = cValue2
			self.button:SetScale(questItemIconSize/self.button.height)
			self.button:SetParent(self)
			self.button:SetPoint("TOPLEFT", self, "TOPLEFT", module.db.profile.questsCellPadding, 0)
			self.button:SetToplevel(true)
			self.button:Show()
		end
	end
	return width, height
end

function qPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	end
	self:EnableMouse(false)
	self.cValue1, self.duration = nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
end

function qtPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cValue1, cValue2)
	local width, height = baseCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	self:EnableMouse(true)
	self.cValue1 = cValue1
	self.duration = 0
	self:SetScript("OnEnter", questOnEnter)
	self:SetScript("OnLeave", questOnLeave)
	self:SetScript("OnMouseDown", questOnMouseDown)
	self:SetScript("OnUpdate", timedQuestUpdate)
	if cValue2 ~= 0 then
		if module.db.profile.showQuestItemIcon and (module.db.profile.hideCompletedObjective == 3 or (module.db.profile.hideCompletedObjective == 2 and core.quests[cValue1].partyComplete == 0) or (module.db.profile.hideCompletedObjective == 1 and core.quests[cValue1].playerComplete == 0)) then
			if core.quests[cValue1].item and core.quests[cValue1].id then
				self.button = core:getItemButton()
				self.button:SetID(core.quests[cValue1].id)
				self.button:SetScript("OnEvent", itemOnEvent)
				local itemCount = GetItemCount(cValue2)
				if itemCount > 1 then
					self.button.count:SetText(itemCount)
					self.button.count:Show()
				end
				if core.quests[cValue1].itemHasRange then
					self.button.rangeTimer = -1
					self.button:SetScript("OnUpdate", updateItemRange)
					self.button:RegisterEvent("PLAYER_TARGET_CHANGED")
				end
				self.button:RegisterEvent("BAG_UPDATE_COOLDOWN")
				local start, duration, enable = GetItemCooldown(cValue2)
				if start > 0 and duration > 0 and enable > 0 then
					self.button.coolDown:SetCooldown(start, duration)
					self.button.coolDown:Show()
				else
					self.button.coolDown:Hide()
				end
			else
				self.button = core:getNormalButton()
				self.button:SetScript("OnEnter", buttonOnEnter)
				self.button:SetScript("OnLeave", buttonOnLeave)
				self.button:SetScript("OnClick", wrongButtonOnClick)
			end
			self.button.iconTexture:SetTexture(core.quests[cValue1].specialItemIcon)
			self.button.link = cValue2
			self.button:SetScale(questItemIconSize/self.button.height)
			self.button:SetParent(self)
			self.button:SetPoint("TOPLEFT", self, "TOPLEFT", questsCellPadding, 0)
			self.button:SetToplevel(true)
			self.button:Show()
		end
	end
	return width, height
end

function qtPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	end
	self:EnableMouse(false)
	self.cValue1, self.duration = nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
	self:SetScript("OnUpdate", nil)
end

function qoPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cValue1)
	local width, height = baseCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	self:EnableMouse(true)
	self.cValue1 = cValue1
	self:SetScript("OnEnter", questOnEnter)
	self:SetScript("OnLeave", questOnLeave)
	self:SetScript("OnMouseDown", questOnMouseDown)
	return width, height
end

function qoPrototype:ReleaseCell()
	self:EnableMouse(false)
	self.cValue1 = nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
end

function qiPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cValue1, cValue2)
	local width, height = baseCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	elseif self.statusBar then
		core:recycleStatusBar(self.statusBar)
		self.statusBar = nil
	end
	self:EnableMouse(true)
	self.cValue1 = cValue1
	self:SetScript("OnMouseDown", questItemOnMouseDown)
	self:SetScript("OnUpdate", nil)
	self.button = core:getNormalButton()
	self.button.iconTexture:SetTexture(cValue2)
	self.button:SetScript("OnEnter", buttonOnEnter)
	self.button:SetScript("OnLeave", buttonOnLeave)
	self.button:SetScript("OnClick", wrongButtonOnClick)
	self.button.link = cValue1
	if module.db.profile.otherItemIconSize >= height then
		self.button:SetScale(height/self.button.height)
	else
		self.button:SetScale(module.db.profile.otherItemIconSize/self.button.height)
	end
	self.button:SetParent(self)
	self.button:SetPoint("TOPLEFT", self, "TOPLEFT", module.db.profile.questsCellPadding, 0)
	self.button:SetToplevel(true)
	self.button:Show()
	return width, height
end

function qiPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	end
	self:EnableMouse(false)
	self.cValue1, self.duration = nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
	self:SetScript("OnUpdate", nil)
end

function module:addQuestToTracker(uid)
	if self.db.char.questsTracked[uid] and core.quests[uid] and ((self.db.profile.hideCompletedQuest == 3) or (self.db.profile.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or (core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + self.db.profile.delayToHideCompletedQuest > self.displayTime) or core.quests[uid].timer)) or (self.db.profile.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or (core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(self.db.profile.delayToHideCompletedQuest) > self.displayTime) or core.quests[uid].timer))) and ((self.db.profile.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not self.db.profile.limitQuestLevel) then
		self.tempLine = self.tracker:AddLine()
		if core.quests[uid].initialTimer then
			core.quests[uid].remainingTime = core.quests[uid].initialTimer + core.quests[uid].startTime - GetTime()
			if core.quests[uid].playersOnQuest > 1 then
				if self.db.profile.hideQuestTag then
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest.."> |cff"..(self.db.profile.colorizeQuestTimer and getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or self.db.profile.defaultQuestTimerColor)..formatTime(core.quests[uid].remainingTime).."|r", qTrFontTQ2, "LEFT", numColumns - 1, qtProvider, questsCellPadding, 0, maxWidth, minWidth, uid, core.quests[uid].specialItemLink or 0)
				else
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeQuestLevel and core.quests[uid].color or self.db.profile.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest.."> |cff"..(self.db.profile.colorizeQuestTimer and getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or self.db.profile.defaultQuestTimerColor)..formatTime(core.quests[uid].remainingTime).."|r", qTrFontTQ2, "LEFT", numColumns - 1, qtProvider, questsCellPadding, 0, maxWidth, minWidth, uid, core.quests[uid].specialItemLink or 0)
				end
			else
				if self.db.profile.hideQuestTag then
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r |cff"..(self.db.profile.colorizeQuestTimer and getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or self.db.profile.defaultQuestTimerColor)..formatTime(core.quests[uid].remainingTime).."|r", qTrFontTQ2, "LEFT", numColumns - 1, qtProvider, questsCellPadding, 0, maxWidth, minWidth, uid, core.quests[uid].specialItemLink or 0)
				else
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeQuestLevel and core.quests[uid].color or self.db.profile.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r |cff"..(self.db.profile.colorizeQuestTimer and getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or self.db.profile.defaultQuestTimerColor)..formatTime(core.quests[uid].remainingTime).."|r", qTrFontTQ2, "LEFT", numColumns - 1, qtProvider, questsCellPadding, 0, maxWidth, minWidth, uid, core.quests[uid].specialItemLink or 0)
				end
			end
		else
			if core.quests[uid].playersOnQuest > 1 then
				if self.db.profile.hideQuestTag then
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest..">", qTrFontTQ2, "LEFT", numColumns - 1, qProvider, questsCellPadding, 0, maxWidth, minWidth, uid, core.quests[uid].specialItemLink or 0)
				else
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeQuestLevel and core.quests[uid].color or self.db.profile.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest..">", qTrFontTQ2, "LEFT", numColumns - 1, qProvider, questsCellPadding, 0, maxWidth, minWidth, uid, core.quests[uid].specialItemLink or 0)
				end
			else
				if self.db.profile.hideQuestTag then
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r", qTrFontTQ2, "LEFT", numColumns - 1, qProvider, questsCellPadding, 0, maxWidth, minWidth, uid, core.quests[uid].specialItemLink or 0)
				else
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeQuestLevel and core.quests[uid].color or self.db.profile.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(self.db.profile.colorizeQuestName and core.quests[uid].color or self.db.profile.defaultQuestNameColor)..core.quests[uid].title.."|r", qTrFontTQ2, "LEFT", numColumns - 1, qProvider, questsCellPadding, 0, maxWidth, minWidth, uid, core.quests[uid].specialItemLink or 0)
				end
			end
		end
		self.tracker:SetCell(self.tempLine, numColumns, core.quests[uid].pqStatus, qTrFontTQ2, "RIGHT", 1, qoProvider, 0, self.db.profile.rightCellPadding, statusMaxWidth, statusMinWidth, uid)
		if core.questsObjectives[uid] then
			for objective, qoData in pairs(core.questsObjectives[uid]) do
				if self.db.profile.hideCompletedObjective == 3 or (self.db.profile.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (self.db.profile.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
					self.tempLine = self.tracker:AddLine()
					if self.db.profile.colorizeQuestObjective then
						self.tracker:SetCell(self.tempLine, 1, "|cff"..qoData.qoColor..qoData.desc.."|r", qoTrFontTQ2, "LEFT", numColumns - 1, qoProvider, objectivesCellPadding, 0, maxWidth, minWidth, uid)
					else
						self.tracker:SetCell(self.tempLine, 1, "|cff"..self.db.profile.defaultQuestObjectiveColor..qoData.desc.."|r", qoTrFontTQ2, "LEFT", numColumns - 1, qoProvider, objectivesCellPadding, 0, maxWidth, minWidth, uid)
					end
					if self.db.profile.colorizeQuestObjectiveCounter then
						self.tracker:SetCell(self.tempLine, numColumns, qoData.qoStatus, qoTrFontTQ2, "RIGHT", 1, qoProvider, 0, self.db.profile.rightCellPadding, statusMaxWidth, statusMinWidth, uid)
					else
						self.tracker:SetCell(self.tempLine, numColumns, qoData.qoStatusNC, qoTrFontTQ2, "RIGHT", 1, qoProvider, 0, self.db.profile.rightCellPadding, statusMaxWidth, statusMinWidth, uid)
					end
				end
			end
		elseif core.quests[uid].objective then
			self.tempLine = self.tracker:AddLine()
			self.tracker:SetCell(self.tempLine, 1, "|cff"..self.db.profile.defaultQuestDescriptionColor..core.quests[uid].objective.."|r", qoTrFontTQ2, "LEFT", numColumns, qoProvider, objectivesCellPadding, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, uid)
		end
		self.tq = true
		self.replaceHeader = nil
	end
end

function module:updateQuests()
	self.tq = nil
	self.replaceHeader = nil
	self.displayTime = time()
	totalMaxWidth = maxWidth + statusMaxWidth + CELL_MARGIN_H * (numColumns - 1)
	totalMinWidth = normalizeWidth(minWidth + statusMinWidth + CELL_MARGIN_H * (numColumns - 1), totalMaxWidth)
	if self.db.profile.displayHeader then
		for _, zone in ipairs(core.sortedZones) do
			if ((self.db.profile.onlyCurrentZone and zone == currentZone) or not self.db.profile.onlyCurrentZone) and core.zones[zone] then
				self.db.char.zoneDisplay[zone] = self.db.char.zoneDisplay[zone] or 1
				if not self.replaceHeader then self.tempLine = self.tracker:AddLine() end
				self.replaceHeader = true
				if not self.db.profile.hideZoneLevelRange then
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeHeaderName and core.zones[zone].color or self.db.profile.defaultHeaderNameColor)..zone.."|r |cff"..(self.db.profile.colorizeHeaderLevel and core.zones[zone].color or self.db.profile.defaultHeaderLevelColor)..core.zones[zone].level.."|r", hTrFontTQ2, "LEFT", numColumns, zProvider, collapseButtonSize, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, zone, self.db.char.zoneDisplay[zone] == 1 and minus or plus)
				else
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(self.db.profile.colorizeHeaderName and core.zones[zone].color or self.db.profile.defaultHeaderNameColor)..zone.."|r", hTrFontTQ2, "LEFT", numColumns, zProvider, collapseButtonSize, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, zone, self.db.char.zoneDisplay[zone] == 1 and minus or plus)
				end
				if self.db.char.zoneDisplay[zone] == 1 or disableInteraction then
					for _, uid in ipairs(core.sortedQuestsByZone[zone]) do
						self:addQuestToTracker(uid)
					end
				else
					for _, uid in ipairs(core.sortedQuestsByZone[zone]) do
						if self.db.char.questsTracked[uid] and core.quests[uid] and ((self.db.profile.hideCompletedQuest == 3) or (self.db.profile.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + self.db.profile.delayToHideCompletedQuest > self.displayTime)) or (self.db.profile.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(self.db.profile.delayToHideCompletedQuest) > self.displayTime))) and ((self.db.profile.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not self.db.profile.limitQuestLevel) then
							self.tq = true
							self.replaceHeader = nil
							break
						end
					end
				end
			end
		end
	else
		for _, uid in ipairs(core.sortedQuests) do
			if self.db.char.questsTracked[uid] and core.quests[uid] and ((self.db.profile.onlyCurrentZone and core.quests[uid].zone == currentZone) or not self.db.profile.onlyCurrentZone) and ((self.db.profile.hideCompletedQuest == 3) or (self.db.profile.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or (core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + self.db.profile.delayToHideCompletedQuest > self.displayTime) or core.quests[uid].timer)) or (self.db.profile.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or (core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(self.db.profile.delayToHideCompletedQuest) > self.displayTime) or core.quests[uid].timer))) and ((self.db.profile.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not self.db.profile.limitQuestLevel) then
				self:addQuestToTracker(uid)
			end
		end
	end
	if self.db.profile.showItemsStartingQuest and core.numItemsStartingQuest > 0 then
		self.tq = true
		local zone = core.LC["Items starting a quest"]
		if not self.replaceHeader or not self.tempLine then
			self.tempLine = self.tracker:AddLine()
		end
		self.db.char.zoneDisplay[zone] = self.db.char.zoneDisplay[zone] or 1
		if self.db.profile.colorizeHeaderName then
			self.tracker:SetCell(self.tempLine, 1, "|cffffffff"..zone.."|r", hTrFontTQ2, "LEFT", numColumns, zProvider, collapseButtonSize, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, zone, self.db.char.zoneDisplay[zone] == 1 and minus or plus)
		else
			self.tracker:SetCell(self.tempLine, 1, "|cff"..self.db.profile.defaultHeaderNameColor..zone.."|r", hTrFontTQ2, "LEFT", numColumns, zProvider, collapseButtonSize, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, zone, self.db.char.zoneDisplay[zone] == 1 and minus or plus)
		end
		if self.db.char.zoneDisplay[zone] == 1 then
			for item, itemData in pairs(core.itemStartingQuest) do
				self.tempLine = self.tracker:AddLine()
				self.tracker:SetCell(self.tempLine, 1, "|cff"..itemData.color..itemData.name.."|r", qTrFontTQ2, "LEFT", numColumns, qiProvider, self.db.profile.questsCellPadding + self.db.profile.otherItemIconSize, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, item, itemData.icon)
			end
		end
		self.replaceHeader = nil
	end
	if self.db.profile.showUsableQuestItemsNotAttached and core.numUsableQuestItems > 0 then
		self.tq = true
		local zone = core.LC["Usable quest items not linked to a quest"]
		if not self.replaceHeader or not self.tempLine then
			self.tempLine = self.tracker:AddLine()
		end
		self.db.char.zoneDisplay[zone] = self.db.char.zoneDisplay[zone] or 1
		if self.db.profile.colorizeHeaderName then
			self.tracker:SetCell(self.tempLine, 1, "|cffffffff"..zone.."|r", hTrFontTQ2, "LEFT", numColumns, zProvider, collapseButtonSize, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, zone, self.db.char.zoneDisplay[zone] == 1 and minus or plus)
		else
			self.tracker:SetCell(self.tempLine, 1, "|cff"..self.db.profile.defaultHeaderNameColor..zone.."|r", hTrFontTQ2, "LEFT", numColumns, zProvider, collapseButtonSize, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, zone, self.db.char.zoneDisplay[zone] == 1 and minus or plus)
		end
		if self.db.char.zoneDisplay[zone] == 1 then
			for item, itemData in pairs(core.usableQuestItems) do
				self.tempLine = self.tracker:AddLine()
				self.tracker:SetCell(self.tempLine, 1, "|cff"..(itemData.color or "ffffff")..itemData.name.."|r", qTrFontTQ2, "LEFT", numColumns, qiProvider, self.db.profile.questsCellPadding + self.db.profile.otherItemIconSize, self.db.profile.rightCellPadding, totalMaxWidth, totalMinWidth, item, itemData.icon)
			end
		end
		self.replaceHeader = nil
	end
	if self.replaceHeader then
		self.tracker:SetCell(self.tempLine, 1, nil, hTrFontTQ2, "LEFT", numColumns)
	end
end

function module:updateQuestsTracker()
	if self.db.profile.toggleTracker and not self.hideInCombat then
		if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
			if achievementTracker.numTrackedAchievement == 0 then
				self.achievementDisplayed = nil
				numColumns = 2
			else
				self.achievementDisplayed = true
				numColumns = achievementTracker.db.profile.numBooleanCriteriaColumns + 1
			end
			self.tracker = QTIP:Acquire("tomQuest2Tracker", numColumns)
			achievementTracker.tracker = self.tracker
			self.tracker:Clear()
			self.tracker:SetBackdrop(backdrop)
			self.tracker:SetBackdropColor(self.db.profile.backDropColor[1], self.db.profile.backDropColor[2], self.db.profile.backDropColor[3], self.db.profile.backDropColor[4])
			self.tracker:SetBackdropBorderColor(self.db.profile.borderColor[1], self.db.profile.borderColor[2], self.db.profile.borderColor[3], self.db.profile.borderColor[4])
			self.tracker:SetScale(self.db.profile.scale)
			self.tracker:SetFrameStrata("BACKGROUND")
			if self.achievementDisplayed and achievementTracker.db.profile.positionInQuestTracker == 1 then
				achievementTracker:updateAchievements()
				self:updateQuests()
			else
				self:updateQuests()
				if self.achievementDisplayed then
					achievementTracker:updateAchievements()
				elseif not self.tq then
					QTIP:Release(self.tracker)
					achievementTracker.tracker = nil
					self.tracker = nil
					return
				end
			end
			local tempLine = self.tracker:AddLine()
			for i = 1, numColumns - 1 do
				self.tracker:SetCell(tempLine, i, " ", fixQTrFontTQ2, "LEFT", 1, nil, 0, 0, maxWidth/(numColumns - 1), minWidth/(numColumns - 1))
			end
			self.tracker:SetCell(tempLine, numColumns, " ", fixQTrFontTQ2, "RIGHT", 1, nil, 0, 0, statusMaxWidth, statusMinWidth)
		else
			numColumns = 2
			self.tracker = QTIP:Acquire("tomQuest2Tracker", 2)
			self.tracker:Clear()
			self.tracker:SetBackdropColor(self.db.profile.backDropColor[1], self.db.profile.backDropColor[2], self.db.profile.backDropColor[3], self.db.profile.backDropColor[4])
			self.tracker:SetBackdropBorderColor(self.db.profile.borderColor[1], self.db.profile.borderColor[2], self.db.profile.borderColor[3], self.db.profile.borderColor[4])
			self.tracker:SetScale(self.db.profile.scale)
			self.tracker:SetFrameStrata("BACKGROUND")
			self:updateQuests()
			if not self.tq then
				QTIP:Release(self.tracker)
				self.tracker = nil
				return
			end
			local tempLine = self.tracker:AddLine()
			self.tracker:SetCell(tempLine, 1, " ", fixQTrFontTQ2, "LEFT", 1, nil, 0, 0, maxWidth, minWidth)
			self.tracker:SetCell(tempLine, 2, " ", fixQTrFontTQ2, "RIGHT", 1, nil, 0, 0, statusMaxWidth, statusMinWidth)
		end
	else
		QTIP:Release(self.tracker)
		self.tracker = nil
		return
	end
	self.tracker:ClearAllPoints()
	if self.db.profile.showAnchor then
		self.tracker:SetPoint(anchorSetPoint[self.db.profile.anchorPosition].point, self.anchor, anchorSetPoint[self.db.profile.anchorPosition].relPoint)
	else
		self.tracker:SetPoint(anchorSetPoint[self.db.profile.anchorPosition].point, self.anchor, anchorSetPoint[self.db.profile.anchorPosition].relPoint, anchorSetPoint[self.db.profile.anchorPosition].x, anchorSetPoint[self.db.profile.anchorPosition].y)
	end
	if not disableInteraction then
		self.tracker:UpdateScrolling(self.db.profile.hasMaxHeight and self.db.profile.maxHeight or nil)
		self.tracker:EnableMouse(true)
	else
		self.tracker:EnableMouse(false)
	end
	self.tracker:Show()
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2QUESTTRACKERTOGGLE = core.LC["Quests Tracker"]

function tQ2toggleQuestTracker(display)
	if not display then
		module.db.profile.toggleTracker = not module.db.profile.toggleTracker
		if not module.db.profile.toggleTracker then
			QTIP:Release(module.tracker)
			module.tracker = nil
		else
			module:updateQuestsTracker()
		end
	elseif display == 1 and not module.db.profile.toggleTracker then
		module.db.profile.toggleTracker = not module.db.profile.toggleTracker
		module:updateQuestsTracker()
	elseif display == 0 and module.db.profile.toggleTracker then
		module.db.profile.toggleTracker = not module.db.profile.toggleTracker
		QTIP:Release(module.tracker)
		module.tracker = nil
	end
end